import sys

from importlib import import_module
from django.conf import settings
from django_huey.exceptions import ConfigurationError
from huey.contrib.djhuey import default_backend_path, get_backend


class DjangoHueySettingsReader:
    def __init__(self, hueys_setting):
        if not isinstance(hueys_setting, dict):
            raise ConfigurationError('Error: DJANGO_HUEY must be a dictionary')
        self.hueys_setting = hueys_setting.get('queues')
        self.default_queue = hueys_setting.get('default')

        if self.default_queue is not None and self.default_queue not in self.hueys_setting:
            raise ConfigurationError(f"Queue defined as default: {self.default_queue}, is not configured in DJANGO_HUEY.")
        self.hueys = {}

    def configure(self):
        new_hueys = dict()
        for queue_name, config in self.hueys_setting.items():
            huey_config = config.copy()

            new_hueys[queue_name] = self._configure_instance(huey_config, queue_name)

        self.hueys_setting = new_hueys


    def get_queue(self, queue):
        return self.hueys_setting[self.get_queue_name(queue)]

    def get_queue_name(self, queue_name):
        if queue_name is None:
            if self.default_queue is None:
                self._raise_queue_config_error()

            queue_name = self.default_queue
        return queue_name

    def _raise_queue_config_error(self):
        raise ConfigurationError("""
Command djangohuey must receive a --queue parameter or define a default queue in DJANGO_HUEY setting.
i.e.: 
python manage.py djangohuey --queue first

or in settings file:

DJANGO_HUEY = {
    'default': 'your-default-queue-name',
    'queues': {
        #Your queues here
    }
}
""")
    def _configure_instance(self, huey_config, default_queue_name):
        name = huey_config.pop('name', default_queue_name)
        if 'backend_class' in huey_config:
            huey_config['huey_class'] = huey_config.pop('backend_class')
        backend_path = huey_config.pop('huey_class', default_backend_path)
        conn_kwargs = huey_config.pop('connection', {})
        try:
            del huey_config['consumer']  # Don't need consumer opts here.
        except KeyError:
            pass
        if 'immediate' not in huey_config:
            huey_config['immediate'] = settings.DEBUG
        huey_config.update(conn_kwargs)

        try:
            backend_cls = get_backend(backend_path)
        except (ValueError, ImportError, AttributeError):
            raise ConfigurationError(f'Error: could not import Huey backend: {backend_path}')

        return backend_cls(name, **huey_config)



import unittest
from unittest import mock
from huey import RedisHuey, MemoryHuey
from django_huey.exceptions import ConfigurationError

from django_huey.config import DjangoHueySettingsReader

class DjangoHueyTests(unittest.TestCase):
    def test_djangohuey_config_with_no_settings(self):

        with self.assertRaises(ConfigurationError) as cm:
            config = DjangoHueySettingsReader(None)

        self.assertEqual('Error: DJANGO_HUEY must be a dictionary', str(cm.exception))


    def test_djangohuey_configure_does_not_raise_error_when_both_settings_are_defined(self):
        DJANGO_HUEY = {
            'queues': {
                'queuename': {
                    'name': 'test',
                    'immediate': True,
                }
            }
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

    def test_djangohuey_get_queue_when_queue_is_none(self):
        DJANGO_HUEY = {
            'queues': {
                'queuename': {
                    'name': 'test',
                    'immediate': True,
                }
            }
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

        with self.assertRaises(ConfigurationError) as cm:
            config.get_queue(None)

        self.assertEqual("""
Command djangohuey must receive a --queue parameter or define a default queue in DJANGO_HUEY setting.
i.e.: 
python manage.py djangohuey --queue first

or in settings file:

DJANGO_HUEY = {
    'default': 'your-default-queue-name',
    'queues': {
        #Your queues here
    }
}
""", str(cm.exception))

    def test_djangohuey_get_queue_when_queue_is_none_and_default_queue_is_defined(self):
        DJANGO_HUEY = {
            'default': 'queuename',
            'queues': {
                'queuename': {
                    'name': 'queue-default',
                    'immediate': True,
                }
            }
        }
        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

        queue = config.get_queue(None)

        self.assertEqual(queue.name, 'queue-default')

    def test_djangohuey_invalid_default_queue(self):
        DJANGO_HUEY = {
            'default': 'invalid-queue',
            'queues': {
                'queuename': {
                    'name': 'queue-default',
                    'immediate': True,
                }
            }
        }
        with self.assertRaises(ConfigurationError) as cm:
            config = DjangoHueySettingsReader(DJANGO_HUEY)

        self.assertEqual("Queue defined as default: invalid-queue, is not configured in DJANGO_HUEY.", str(cm.exception))

    def test_djangohuey_configure_when_django_huey_setting_is_defined(self, *args):
        DJANGO_HUEY = {
            'queues': {
                'first': {
                    'huey_class': 'huey.RedisHuey',  # Huey implementation to use.
                    'name': 'testname',  # Use db name for huey.
                },
                'mails': {
                    'huey_class': 'huey.MemoryHuey',  # Huey implementation to use.
                    'name': 'testnamememory',  # Use db name for huey.
                }
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()

        self.assertTrue(isinstance(config.get_queue('first'), RedisHuey))
        self.assertEqual(config.get_queue('first').name, 'testname')
        self.assertTrue(isinstance(config.get_queue('mails'), MemoryHuey))
        self.assertEqual(config.get_queue('mails').name, 'testnamememory')

    def test_djangohuey_configure_when_django_huey_setting_is_an_object_raises_error(self, *args):
        DJANGO_HUEY = object()

        with self.assertRaises(ConfigurationError) as cm:
            config = DjangoHueySettingsReader(DJANGO_HUEY)

        self.assertEqual('Error: DJANGO_HUEY must be a dictionary', str(cm.exception))

    def test_djangohuey_if_name_is_not_defined_queue_name_is_default(self, *args):
        DJANGO_HUEY = {
            'queues': {
                'first': {
                    'huey_class': 'huey.RedisHuey',  # Huey implementation to use.
                },
                'mails': {
                    'huey_class': 'huey.MemoryHuey',  # Huey implementation to use.
                }
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()
        self.assertEqual(config.get_queue('first').name, 'first')

        self.assertEqual(config.get_queue('mails').name, 'mails')


    def test_djangohuey_with_backend_class(self, *args):
        DJANGO_HUEY = {
            'queues': {
                'first': {
                    'backend_class': 'huey.RedisHuey',
                }
            }
        }

        config = DjangoHueySettingsReader(DJANGO_HUEY)

        config.configure()
        self.assertTrue(isinstance(config.get_queue('first'), RedisHuey))

    def test_djangohuey_invalid_backend_class(self, *args):
        DJANGO_HUEY = {
            'queues': {
                'first': {
                    'backend_class': 'huey.RedisHuey2',
                }
            }
        }

        with self.assertRaises(ConfigurationError) as cm:
            config = DjangoHueySettingsReader(DJANGO_HUEY)
            config.configure()

        self.assertEqual('Error: could not import Huey backend: huey.RedisHuey2', str(cm.exception))



DEBUG = True

DATABASES = {
    "default": {
        'NAME': 'testdatabase',
        "ENGINE": "django.db.backends.sqlite3"
    }
}


SECRET_KEY = "django_tests_secret_key"
TIME_ZONE = "America/Chicago"
LANGUAGE_CODE = "en-us"
ADMIN_MEDIA_PREFIX = "/static/admin/"
STATICFILES_DIRS = ()

MIDDLEWARE_CLASSES = []

INSTALLED_APPS = [
    "django_huey"
]

DJANGO_HUEY = {
    'default': 'multi-2',
    'queues': {
        'multi-1': {
            'huey_class': 'huey.MemoryHuey',  # Huey implementation to use.
            'results': True,  # Store return values of tasks.
            'store_none': False,  # If a task returns None, do not save to results.
            'immediate': False,  # If DEBUG=True, run synchronously.
            'utc': True,  # Use UTC for all times internally.
            'blocking': True,  # Perform blocking pop rather than poll Redis.
            'consumer': {
                'workers': 1,
                'worker_type': 'thread',
                'initial_delay': 0.1,  # Smallest polling interval, same as -d.
                'backoff': 1.15,  # Exponential backoff using this rate, -b.
                'max_delay': 10.0,  # Max possible polling interval, -m.
                'scheduler_interval': 60,  # Check schedule every second, -s.
                'periodic': True,  # Enable crontab feature.
                'check_worker_health': True,  # Enable worker health checks.
                'health_check_interval': 300,  # Check worker health every second.
            },
        },
        'multi-2': {
            'huey_class': 'huey.MemoryHuey',  # Huey implementation to use.
            'results': True,  # Store return values of tasks.
            'store_none': False,  # If a task returns None, do not save to results.
            'immediate': False,  # If DEBUG=True, run synchronously.
            'utc': True,  # Use UTC for all times internally.
            'blocking': True,  # Perform blocking pop rather than poll Redis.
            'consumer': {
                'workers': 1,
                'worker_type': 'thread',
                'initial_delay': 0.1,  # Smallest polling interval, same as -d.
                'backoff': 1.15,  # Exponential backoff using this rate, -b.
                'max_delay': 10.0,  # Max possible polling interval, -m.
                'scheduler_interval': 60,  # Check schedule every second, -s.
                'periodic': True,  # Enable crontab feature.
                'check_worker_health': True,  # Enable worker health checks.
                'health_check_interval': 300,  # Check worker health every second.
            },
        }
    }
}
HUEY = {
    'name': 'test',
    'immediate': True,
}



from functools import wraps
from importlib import import_module
import sys
import traceback

from django.conf import settings
from django.db import close_old_connections

from django_huey.config import DjangoHueySettingsReader

DJANGO_HUEY = getattr(settings, 'DJANGO_HUEY', None)

config = DjangoHueySettingsReader(DJANGO_HUEY)
config.configure()

def get_close_db_for_queue(queue):
    def close_db(fn):
        """Decorator to be used with tasks that may operate on the database."""
        @wraps(fn)
        def inner(*args, **kwargs):
            instance = get_queue(queue)
            if not instance.immediate:
                close_old_connections()
            try:
                return fn(*args, **kwargs)
            finally:
                if not instance.immediate:
                    close_old_connections()
        return inner

    return close_db

def get_queue(queue):
    return config.get_queue(queue)

def get_queue_name(queue):
    return config.get_queue_name(queue)

def task(*args, queue=None, **kwargs):
    return get_queue(queue).task(*args, **kwargs)

def periodic_task(*args, queue=None, **kwargs):
    return get_queue(queue).periodic_task(*args, **kwargs)
    
def lock_task(*args, queue=None, **kwargs):
    return get_queue(queue).lock_task(*args, **kwargs)
    
# Task management.
    
def enqueue(*args, queue=None, **kwargs):
    return get_queue(queue).enqueue(*args, **kwargs)
    
def restore(*args, queue=None, **kwargs):
    return get_queue(queue).restore(*args, **kwargs)
    
def restore_all(*args, queue=None, **kwargs):
    return get_queue(queue).restore_all(*args, **kwargs)
    
def restore_by_id(*args, queue=None, **kwargs):
    return get_queue(queue).restore_by_id(*args, **kwargs)
    
def revoke(*args, queue=None, **kwargs):
    return get_queue(queue).revoke(*args, **kwargs)
    
def revoke_all(*args, queue=None, **kwargs):
    return get_queue(queue).revoke_all(*args, **kwargs)
    
def revoke_by_id(*args, queue=None, **kwargs):
    return get_queue(queue).revoke_by_id(*args, **kwargs)
    
def is_revoked(*args, queue=None, **kwargs):
    return get_queue(queue).is_revoked(*args, **kwargs)
    
def result(*args, queue=None, **kwargs):
    return get_queue(queue).result(*args, **kwargs)
    
def scheduled(*args, queue=None, **kwargs):
    return get_queue(queue).scheduled(*args, **kwargs)
    
# Hooks.
def on_startup(*args, queue=None, **kwargs):
    return get_queue(queue).on_startup(*args, **kwargs)
    
def on_shutdown(*args, queue=None, **kwargs):
    return get_queue(queue).on_shutdown(*args, **kwargs)
    
def pre_execute(*args, queue=None, **kwargs):
    return get_queue(queue).pre_execute(*args, **kwargs)
    
def post_execute(*args, queue=None, **kwargs):
    return get_queue(queue).post_execute(*args, **kwargs)
    
def signal(*args, queue=None, **kwargs):
    return get_queue(queue).signal(*args, **kwargs)
    
def disconnect_signal(*args, queue=None, **kwargs):
    return get_queue(queue).disconnect_signal(*args, **kwargs)


def db_task(*args, **kwargs):
    queue = kwargs.get('queue')
    def decorator(fn):
        ret = task(*args, **kwargs)(get_close_db_for_queue(queue)(fn))
        ret.call_local = fn
        return ret
    return decorator

def db_periodic_task(*args, **kwargs):
    queue = kwargs.get('queue')
    def decorator(fn):
        ret = periodic_task(*args, **kwargs)(get_close_db_for_queue(queue)(fn))
        ret.call_local = fn
        return ret
    return decorator



import unittest

from unittest import mock
from django_huey import task, periodic_task, lock_task, enqueue, restore, restore_all, restore_by_id, revoke\
,revoke_all, revoke_by_id, is_revoked, result, scheduled, on_startup, on_shutdown, pre_execute, post_execute\
,signal, disconnect_signal 

DECORATORS = [task, periodic_task, lock_task, enqueue, restore, restore_all, restore_by_id, revoke\
,revoke_all, revoke_by_id, is_revoked, result, scheduled, on_startup, on_shutdown, pre_execute, post_execute\
,signal, disconnect_signal]


class DecoratorsTest(unittest.TestCase):
    def make_test_decorator(self, decorator):
        with mock.patch('django_huey.get_queue') as obj:
            @decorator(queue='test_queue')
            def test_func():
                pass
            
            obj.assert_called_once_with("test_queue")

    def test_decorators(self):
        for decorator in DECORATORS:
            self.make_test_decorator(decorator)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="django-huey",
    version="1.0.1",
    scripts=[],
    author="GAIA - Cooperativa de desarrollo de software",
    author_email="contacto@gaiacoop.tech",
    description="An extension for django and huey that supports multi queue management",
    long_description=long_description,
    install_requires=[
        "django>=3.2",
        "huey>=2.0",
    ],
    long_description_content_type="text/markdown",
    url="https://github.com/gaiacoop/django-huey",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)